import os
import re
from typing import Tuple, List, Any

import thingsvision.vision 
from thingsvision.dataloader import DataLoader
from thingsvision.dataset import ImageDataset
from thingsvision.model_class import Model

import numpy as np

import skimage
import imageio

import torch 
import torch.nn as nn

import tensorflow as tf
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense

DATA_PATH = './data'
TEST_PATH = './test_images'
OUT_PATH = './test'

PT_MODEL_AND_MODULE_NAMES = {
    'vgg16_bn': ['features.23', 'classifier.3'],
    'vgg19_bn': ['features.23', 'classifier.3'],
    'cornet_r': ['decoder.flatten'],
    'cornet_rt': ['decoder.flatten'],
    'cornet_s': ['decoder.flatten'],
    'cornet_z': ['decoder.flatten'],
    'clip-ViT': ['visual'],
    'clip-RN': ['visual'],
    'VGG16bn_ecoset': ['classifier.3']
}

TF_MODEL_AND_MODULES_NAMES = {
    'VGG16': ['block1_conv1', 'flatten'],
    'VGG19': ['block1_conv1', 'flatten']
}

BACKENDS = ['tf', 'pt']

FILE_FORMATS = ['hdf5', 'npy', 'mat', 'txt']
DISTANCES = ['correlation', 'cosine', 'euclidean', 'gaussian']

BATCH_SIZE = 16
NUM_OBJECTS = 1854
# we want to iterate over two batches to exhaustively test mini-batching
NUM_SAMPLES = int(BATCH_SIZE * 2)
DEVICE = 'cuda' if torch.cuda.is_available() else 'cpu'


tf_model = Sequential()
tf_model.add(Dense(2, input_dim=1, activation='relu',
             use_bias=False, name='relu'))
weights = np.array([[[1,1]]])
tf_model.get_layer('relu').set_weights(weights)
tf_model.add(Dense(2, input_dim=2, activation='relu',
             use_bias=False, name='relu2'))
weights = np.array([[[1,1], [1,1]]])
tf_model.get_layer('relu2').set_weights(weights)


class NN(nn.Module):

    def __init__(self, in_size: int, out_size: int):
        super(NN, self).__init__()
        self.linear = nn.Linear(in_size, out_size, bias=False)
        self.relu = nn.ReLU()
        self.linear2 = nn.Linear(out_size, 1, bias=False)
        # exchange weight value with 1.
        self.linear.weight = nn.Parameter(torch.tensor([[1.], [1.]]))
        self.linear2.weight = nn.Parameter(torch.tensor([[1., 1.], [1., 1.]]))
        self.relu2 = nn.ReLU()

    def forward(self, x):
        x = self.linear(x)
        act = self.relu(x)
        print(act)
        y = self.linear2(act)
        act = self.relu2(y)
        print(y)
        return y


pt_model = NN(1, 2)


class SimpleDataset(object):

    def __init__(self, values: List[int], backend: str):
        self.values = values
        self.backend = backend

    def __getitem__(self, idx: int) -> Tuple[Any, Any]:
        target = 0
        value = self.values[idx]

        if self.backend == 'pt':
            value = torch.tensor([float(value)])
            target = torch.tensor([target])
        elif self.backend == 'tf':
            value = tf.convert_to_tensor([float(value)])
            target = tf.convert_to_tensor([target])

        return value, target

    def __len__(self) -> int:
        return len(self.values)

def iterate_through_all_model_combinations():
    for backend in BACKENDS:
        MODEL_AND_MODULE_NAMES = None
        if backend == 'pt':
            MODEL_AND_MODULE_NAMES = PT_MODEL_AND_MODULE_NAMES
        elif backend == 'tf':
            MODEL_AND_MODULE_NAMES = TF_MODEL_AND_MODULES_NAMES

        for model_name in MODEL_AND_MODULE_NAMES:
            model, dataset, dl = create_model_and_dl(model_name, backend)
            yield model, dataset, dl, MODEL_AND_MODULE_NAMES[model_name], model_name


def create_model_and_dl(model_name, backend):
    """Iterate through all backends and models and create model, dataset and data loader."""
    model = Model(
        model_name=model_name,
        pretrained=True,
        device=DEVICE,
        backend=backend
    )

    dataset = ImageDataset(
        root=TEST_PATH,
        out_path=OUT_PATH,
        backend=backend,
        imagenet_train=None,
        imagenet_val=None,
        things=None,
        things_behavior=None,
        add_ref_imgs=None,
        transforms=model.get_transformations()
    )
    dl = DataLoader(
        dataset,
        batch_size=BATCH_SIZE,
        backend=backend,
    )
    return model, dataset, dl

def create_test_images(n_samples: int = NUM_SAMPLES) -> None:
    """Create an artificial image dataset to be used for performing tests."""
    if not os.path.exists(OUT_PATH):
        os.makedirs(OUT_PATH)
    if not os.path.exists(TEST_PATH):
        test_img_1 = skimage.data.hubble_deep_field()
        test_img_2 = skimage.data.coffee()
        test_imgs = list(map(lambda x: x / x.max(), [test_img_1, test_img_2]))

        classes = ['hubble', 'coffee']
        for cls in classes:
            PATH = os.path.join(TEST_PATH, cls)
            if not os.path.exists(PATH):
                os.makedirs(PATH)

        for i in range(n_samples):
            if i > n_samples // 2:
                test_img = np.copy(test_imgs[0])
                cls = classes[0]
            else:
                test_img = np.copy(test_imgs[1])
                cls = classes[1]
            H, W, C = test_img.shape
            # add random Gaussian noise to test image
            noisy_img = test_img + np.random.randn(H, W, C)
            noisy_img = noisy_img.astype(np.uint8)
            imageio.imsave(os.path.join(
                TEST_PATH, cls, f'test_img_{i+1:03d}.png'), noisy_img)
        print('\n...Successfully created image dataset for testing.\n')

def download_item_names():
    if not os.path.isfile(os.path.join(DATA_PATH, 'item_names.tsv')):
        try:
            os.system(
                "curl -O https://raw.githubusercontent.com/ViCCo-Group/THINGSvision/master/get_files.sh")
            os.system("bash get_files.sh")
        except:
            raise RuntimeError(
                "Download the THINGS item names <tsv> file to run test;\n"
                "See README.md for details."
            )



import os
import re
import unittest

import tests.helper as helper 
import thingsvision.vision as vision
from thingsvision.dataloader import DataLoader
from thingsvision.model_class import Model

import numpy as np 

class ExtractionPretrainedTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        helper.create_test_images()
        
    def test_extraction_pretrained_modells(self):
        """Tests basic feature extraction pipeline."""
        for model, dataset, dl, module_names, model_name in helper.iterate_through_all_model_combinations():
            clip = re.search(r'^clip', model_name)
            self.assertEqual(len(dataset), len(dl) * helper.BATCH_SIZE)
            amount_objects = len(dataset)

            for module_name in module_names:
                features, targets = model.extract_features(
                        data_loader=dl,
                        module_name=module_name,
                        flatten_acts=False,
                        clip=clip,
                        return_probabilities=False
                )

                self.assertTrue(isinstance(features, np.ndarray))
                self.assertTrue(isinstance(targets, np.ndarray))
                self.assertEqual(features.shape[0], amount_objects)
                self.assertEqual(len(targets), amount_objects)

                if module_name.startswith('classifier'):
                    self.assertEqual(features.shape[1],
                                     model.model.classifier[int(module_name[-1])].out_features)

                if not clip:
                    features, targets, probs = model.extract_features(
                        data_loader=dl,
                        module_name=module_name,
                        flatten_acts=False,
                        clip=clip,
                        return_probabilities=True
                    )

                    self.assertTrue(isinstance(features, np.ndarray))
                    self.assertTrue(isinstance(targets, np.ndarray))
                    self.assertEqual(features.shape[0], amount_objects)
                    self.assertEqual(len(targets), amount_objects)
                    self.assertEqual(len(probs), amount_objects)

    def test_extraction_across_models(self):
        backend = 'pt'
        model_names = ['vgg16_bn', 'vgg19_bn']
        module_names = ['features.23', 'features.23']
        img_paths = [helper.TEST_PATH, helper.TEST_PATH]
        vision.extract_features_across_models_and_datasets(
            out_path=helper.OUT_PATH,
            model_names=model_names,
            img_paths=img_paths,
            module_names=module_names,
            clip=[False, False],
            pretrained=True,
            batch_size=1,
            backend=backend,
            flatten_acts=False,
            )
        path1 = os.path.join(helper.OUT_PATH, helper.TEST_PATH, model_names[0],
                                module_names[0], 'features')
        self.assertTrue(os.path.exists(path1))
        path2 = os.path.join(helper.OUT_PATH, helper.TEST_PATH, model_names[1],
                                module_names[1], 'features')
        self.assertTrue(os.path.exists(path2))

    def test_extraction_across_models_and_modules(self):
        backend = 'pt'
        model_names = ['vgg16_bn', 'alexnet']
        module_names = ['features.22', 'features.1']
        img_paths = [helper.TEST_PATH, helper.TEST_PATH]
        vision.extract_features_across_models_datasets_and_modules(
            out_path=helper.OUT_PATH,
            model_names=model_names,
            img_paths=img_paths,
            module_names=module_names,
            clip=[False, False],
            pretrained=False,
            batch_size=1,
            backend=backend, 
            flatten_acts=True,
            )
        path1 = os.path.join(helper.OUT_PATH, helper.TEST_PATH, model_names[0],
                                module_names[0], 'features')
        self.assertTrue(os.path.exists(path1))
        path2 = os.path.join(helper.OUT_PATH, helper.TEST_PATH, model_names[1],
                                module_names[1], 'features')
        self.assertTrue(os.path.exists(path2))

    def test_extraction_across_models_and_modules(self):
        backend = 'tf'
        model_names = ['VGG16', 'VGG19']
        module_names = ['block1_conv1', 'fc1']
        img_paths = [helper.TEST_PATH, helper.TEST_PATH]
        vision.extract_features_across_models_datasets_and_modules(
            out_path=helper.OUT_PATH,
            model_names=model_names,
            img_paths=img_paths,
            module_names=module_names,
            clip=[False, False],
            pretrained=False,
            batch_size=1,
            backend=backend, 
            flatten_acts=True,
            )
        path1 = os.path.join(helper.OUT_PATH, helper.TEST_PATH, model_names[0],
                                module_names[0], 'features')
        self.assertTrue(os.path.exists(path1))
        path2 = os.path.join(helper.OUT_PATH, helper.TEST_PATH, model_names[1],
                                module_names[1], 'features')
        self.assertTrue(os.path.exists(path2))
        
       


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import torch
import torch.utils.model_zoo

from thingsvision.cornet.cornet_z import CORnet_Z
from thingsvision.cornet.cornet_z import HASH as HASH_Z
from thingsvision.cornet.cornet_r import CORnet_R
from thingsvision.cornet.cornet_r import HASH as HASH_R
from thingsvision.cornet.cornet_rt import CORnet_RT
from thingsvision.cornet.cornet_rt import HASH as HASH_RT
from thingsvision.cornet.cornet_s import CORnet_S
from thingsvision.cornet.cornet_s import HASH as HASH_S


def get_model(model_letter, pretrained=False, map_location=None, **kwargs):
    model_letter = model_letter.upper()
    model_hash = globals()[f'HASH_{model_letter}']
    model = globals()[f'CORnet_{model_letter}'](**kwargs)
    model = torch.nn.DataParallel(model)
    if pretrained:
        url = f'https://s3.amazonaws.com/cornet-models/cornet_{model_letter.lower()}-{model_hash}.pth'
        ckpt_data = torch.utils.model_zoo.load_url(url, map_location=map_location)
        model.load_state_dict(ckpt_data['state_dict'])
    return model


def cornet_z(pretrained=False, map_location=None):
    return get_model('z', pretrained=pretrained, map_location=map_location)


def cornet_r(pretrained=False, map_location=None, times=5):
    return get_model('r', pretrained=pretrained, map_location=map_location, times=times)


def cornet_rt(pretrained=False, map_location=None, times=5):
    return get_model('rt', pretrained=pretrained, map_location=map_location, times=times)


def cornet_s(pretrained=False, map_location=None):
    return get_model('s', pretrained=pretrained, map_location=map_location)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import math
from collections import OrderedDict
from torch import nn


HASH = '1d3f7974'


class Flatten(nn.Module):

    """
    Helper module for flattening input tensor to 1-D for the use in Linear modules
    """

    def forward(self, x):
        return x.view(x.size(0), -1)


class Identity(nn.Module):

    """
    Helper module that stores the current tensor. Useful for accessing by name
    """

    def forward(self, x):
        return x


class CORblock_S(nn.Module):

    scale = 4  # scale of the bottleneck convolution channels

    def __init__(self, in_channels, out_channels, times=1):
        super().__init__()

        self.times = times

        self.conv_input = nn.Conv2d(in_channels, out_channels, kernel_size=1, bias=False)
        self.skip = nn.Conv2d(out_channels, out_channels,
                              kernel_size=1, stride=2, bias=False)
        self.norm_skip = nn.BatchNorm2d(out_channels)

        self.conv1 = nn.Conv2d(out_channels, out_channels * self.scale,
                               kernel_size=1, bias=False)
        self.nonlin1 = nn.ReLU(inplace=True)

        self.conv2 = nn.Conv2d(out_channels * self.scale, out_channels * self.scale,
                               kernel_size=3, stride=2, padding=1, bias=False)
        self.nonlin2 = nn.ReLU(inplace=True)

        self.conv3 = nn.Conv2d(out_channels * self.scale, out_channels,
                               kernel_size=1, bias=False)
        self.nonlin3 = nn.ReLU(inplace=True)

        self.output = Identity()  # for an easy access to this block's output

        # need BatchNorm for each time step for training to work well
        for t in range(self.times):
            setattr(self, f'norm1_{t}', nn.BatchNorm2d(out_channels * self.scale))
            setattr(self, f'norm2_{t}', nn.BatchNorm2d(out_channels * self.scale))
            setattr(self, f'norm3_{t}', nn.BatchNorm2d(out_channels))

    def forward(self, inp):
        x = self.conv_input(inp)

        for t in range(self.times):
            if t == 0:
                skip = self.norm_skip(self.skip(x))
                self.conv2.stride = (2, 2)
            else:
                skip = x
                self.conv2.stride = (1, 1)

            x = self.conv1(x)
            x = getattr(self, f'norm1_{t}')(x)
            x = self.nonlin1(x)

            x = self.conv2(x)
            x = getattr(self, f'norm2_{t}')(x)
            x = self.nonlin2(x)

            x = self.conv3(x)
            x = getattr(self, f'norm3_{t}')(x)

            x += skip
            x = self.nonlin3(x)
            output = self.output(x)

        return output


def CORnet_S():
    model = nn.Sequential(OrderedDict([
        ('V1', nn.Sequential(OrderedDict([  # this one is custom to save GPU memory
            ('conv1', nn.Conv2d(3, 64, kernel_size=7, stride=2, padding=3,
                            bias=False)),
            ('norm1', nn.BatchNorm2d(64)),
            ('nonlin1', nn.ReLU(inplace=True)),
            ('pool', nn.MaxPool2d(kernel_size=3, stride=2, padding=1)),
            ('conv2', nn.Conv2d(64, 64, kernel_size=3, stride=1, padding=1,
                            bias=False)),
            ('norm2', nn.BatchNorm2d(64)),
            ('nonlin2', nn.ReLU(inplace=True)),
            ('output', Identity())
        ]))),
        ('V2', CORblock_S(64, 128, times=2)),
        ('V4', CORblock_S(128, 256, times=4)),
        ('IT', CORblock_S(256, 512, times=2)),
        ('decoder', nn.Sequential(OrderedDict([
            ('avgpool', nn.AdaptiveAvgPool2d(1)),
            ('flatten', Flatten()),
            ('linear', nn.Linear(512, 1000)),
            ('output', Identity())
        ])))
    ]))

    # weight initialization
    for m in model.modules():
        if isinstance(m, nn.Conv2d):
            n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            m.weight.data.normal_(0, math.sqrt(2. / n))
        # nn.Linear is missing here because I originally forgot
        # to add it during the training of this network
        elif isinstance(m, nn.BatchNorm2d):
            m.weight.data.fill_(1)
            m.bias.data.zero_()

    return model



import re
from typing import Tuple, List, Iterator, Dict, Any

import numpy as np 

from PIL import Image
from numpy.core.fromnumeric import resize

import tensorflow as tf
from tensorflow import keras
import tensorflow.keras.applications as tensorflow_models
from tensorflow.keras import layers

import thingsvision.cornet as cornet
import thingsvision.clip as clip
import thingsvision.custom_models as custom_models

import torch
import torch.nn as nn
import torch.nn.functional as F
from torchvision import transforms as T
import torchvision.models as torchvision_models

class Model():
    def __init__(self, 
                 model_name: str, 
                 pretrained: bool,
                 device: str,
                 model_path: str=None,
                 backend: str='pt'):
        """
        Parameters
        ----------
        Model_name : str
            Model name. Name of model for which features should
            subsequently be extracted.
        pretrained : bool
            Whether to load a model with pretrained or
            randomly initialized weights into memory.
        device : str
            Device. Whether model weights should be moved
            to CUDA or left on the CPU.
        model_path : str (optional)
            path/to/weights. If pretrained is set to False,
            model weights can be loaded from a path on the
            user's machine. This is useful when operating
            on a server without network access, or when
            features should be extracted for a model that
            was fine-tuned (or trained) on a custom image
            dataset.
        backend : str (optional)
            Deep learning framework that should be used.
            'pt' for PyTorch and 'tf' for Tensorflow
        """

        self.model_name = model_name
        self.backend = backend
        self.pretrained = pretrained
        self.device = device
        self.model_path = model_path
        self.load_model()


    def load_model(self) -> Tuple[Any, Any]:
        """Load a pretrained *torchvision* or CLIP model into memory."""
        if self.backend == 'pt':
            if re.search(r'^clip', self.model_name):
                clip_model_name = "RN50"
                if re.search(r'ViT$', self.model_name):
                    clip_model_name = "ViT-B/32"
                self.model, self.clip_n_px = clip.load(
                                                    clip_model_name,
                                                    device=self.device,
                                                    model_path=self.model_path,
                                                    pretrained=self.pretrained,
                                                    jit=False,
                                            )
            else:
                device = torch.device(self.device)
                if re.search(r'^cornet', self.model_name):
                    try:
                        self.model = getattr(cornet, f'cornet_{self.model_name[-1]}')
                    except:
                        self.model = getattr(cornet, f'cornet_{self.model_name[-2:]}')
                    self.model = self.model(pretrained=self.pretrained, map_location=device)
                    self.model = self.model.module    # remove DataParallel
                elif hasattr(custom_models, self.model_name):
                    self.model = getattr(custom_models, self.model_name)(self.device, self.backend).create_model()
                else:
                    self.model = getattr(torchvision_models, self.model_name)
                    self.model = self.model(pretrained=self.pretrained)
                self.model = self.model.to(device)
            if self.model_path:
                try:
                    state_dict = torch.load(self.model_path, map_location=device)
                except FileNotFoundError:
                    state_dict = torch.hub.load_state_dict_from_url(self.model_path, map_location=device)
                self.model.load_state_dict(state_dict)
            self.model.eval()
        elif self.backend == 'tf':
            if hasattr(custom_models, self.model_name):
                self.model = getattr(custom_models, self.model_name)(self.device, self.backend).create_model()
            else:
                model = getattr(tensorflow_models, self.model_name)
                if self.pretrained:
                    weights = 'imagenet'
                elif self.model_path:
                    weights = self.model_path
                else:
                    weights = None
                
                self.model = model(weights=weights)
    

    def show(self) -> str:
        """Show architecture of model to select a layer."""
        if self.backend == 'pt':
            if re.search(r'^clip', self.model_name):
                for l, (n, p) in enumerate(self.model.named_modules()):
                    if l > 1:
                        if re.search(r'^visual', n):
                            print(n)
                print('visual')
            else:
                print(self.model)
        else:
            print(self.model.summary())
        print(f'\nEnter module name for which you would like to extract features:\n')
        module_name = str(input())
        print()
        return module_name


    def tf_extraction(
        self,
        data_loader: Iterator,
        module_name: str,
        flatten_acts: bool,
        return_probabilities: bool=False,
    ) -> tuple:
        """Feature extraction helper function for TensorFlow models."""
        features, targets = [], []
        if return_probabilities:
            probabilities = []
        for batch in data_loader:
            X, y = batch
            layer_out = [self.model.get_layer(module_name).output]
            activation_model = keras.models.Model(
                inputs=self.model.input,
                outputs=layer_out,
            )
            activations = activation_model.predict(X)
            if flatten_acts:
                activations = activations.reshape(activations.shape[0], -1)
            features.append(activations)
            targets.extend(y.numpy())
            if return_probabilities:
                out = self.model.predict(X)
                probas = tf.nn.softmax(out, axis=1)
                probabilities.append(probas)
        features = np.vstack(features)
        targets = np.asarray(targets).ravel()
        if return_probabilities:
            probabilities = np.vstack(probabilities)
            return features, targets, probabilities
        return features, targets

    
    def pt_extraction(
            self,
            data_loader: Iterator,
            module_name: str,
            flatten_acts: bool,
            clip: bool=False,
            return_probabilities: bool=False,
        ) -> tuple:
        """Feature extraction helper function for PyTorch models."""
        if re.search(r'ensemble$', module_name):
            ensembles = ['conv_ensemble', 'maxpool_ensemble']
            assert module_name in ensembles, f'\nIf aggregating filters across layers and subsequently concatenating features, module name must be one of {ensembles}\n'
            if re.search(r'^conv', module_name):
                feature_extractor = nn.Conv2d
            else:
                feature_extractor = nn.MaxPool2d
        device = torch.device(self.device)
        # initialise dictionary to store hidden unit activations per mini-batch
        global activations
        activations = {}
        # register forward hook to store activations
        model = self.register_hook()
        features, targets = [], []
        if return_probabilities:
            probabilities = []
        with torch.no_grad():
            for batch in data_loader:
                X, y = (t.to(device) for t in batch)
                if clip:
                    img_features = model.encode_image(X)
                    if module_name == 'visual':
                        assert torch.unique(activations[module_name] == img_features).item(
                        ), '\nImage features should represent activations in last encoder layer.\n'
                else:
                    out = model(X)
                    if return_probabilities:
                        probas = F.softmax(out, dim=1)
                        probabilities.append(probas)
                if re.compile(r'ensemble$').search(module_name):
                    layers = self.enumerate_layers(feature_extractor)
                    act = self.ensemble_featmaps(activations, layers, 'max')
                else:
                    act = activations[module_name]
                    if flatten_acts:
                        if clip:
                            if re.compile(r'attn$').search(module_name):
                                if isinstance(act, tuple):
                                    act = act[0]
                            else:
                                if act.size(0) != X.shape[0] and len(act.shape) == 3:
                                    act = act.permute(1, 0, 2)
                        act = act.view(act.size(0), -1)
                features.append(act.cpu().numpy())
                targets.extend(y.squeeze(-1).cpu().numpy())
        features = np.vstack(features)
        targets = np.asarray(targets).ravel()
        if return_probabilities:
            probabilities = np.vstack(probabilities)
            return features, targets, probabilities
        return features, targets


    def extract_features(
            self,
            data_loader: Iterator,
            module_name: str,
            flatten_acts: bool,
            clip: bool = False,
            return_probabilities: bool = False,
        ) -> Tuple[np.ndarray, np.ndarray]:
        """Extract hidden unit activations (at specified layer) for every image in the database.

            Parameters
            ----------
            data_loader : Iterator
                Mini-batches. Iterator with equally sized
                mini-batches, where each element is a
                subsample of the full image dataset.
            module_name : str
                Layer name. Name of neural network layer for
                which features should be extraced.
            flatten_acts : bool
                Whether activation tensor (e.g., activations
                from an early layer of the neural network model)
                should be transformed into a vector.
            clip : bool (optional)
                Whether neural network model is a CNN-based
                torchvision or CLIP-based model. Since CLIP
                has a different training objective, feature
                extraction must be performed differently.
                For PyTorch only.
            return_probabilities : bool (optional)
                Whether class probabilities (softmax predictions)
                should be returned in addition to the feature matrix
                and the target vector.

            Returns
            -------
            output : Tuple[np.ndarray, np.ndarray] OR Tuple[np.ndarray, np.ndarray, np.ndarray]
                Returns the feature matrix and the target vector OR in addition to the feature
                matrix and the target vector, the class probabilities.
        """
        if self.backend == 'pt':
            if return_probabilities:
                assert not clip, '\nCannot extract features for CLIP and return class predictions simultaneously. We will add this possibility in a future version.\n'
                features, targets, probabilities = self.pt_extraction(
                    data_loader=data_loader,
                    module_name=module_name,
                    flatten_acts=flatten_acts,
                    clip=clip,
                    return_probabilities=True,
                )
            else:
                features, targets = self.pt_extraction(
                    data_loader=data_loader,
                    module_name=module_name,
                    flatten_acts=flatten_acts,
                    clip=clip, 
                )
        else:
            if return_probabilities:
                features, targets, probabilities = self.tf_extraction(
                    data_loader=data_loader,
                    module_name=module_name,
                    flatten_acts=flatten_acts,
                    return_probabilities=True,
                )
            else:
                features, targets = self.tf_extraction(
                    data_loader=data_loader,
                    module_name=module_name,
                    flatten_acts=flatten_acts,
                )        
        if return_probabilities:
            assert features.shape[0] == len(targets) == len(
                    probabilities), '\nFeatures, targets, and probabilities must correspond to the same number of images.\n'
            print(f'...Features successfully extracted for all {len(features)} images in the database.')
            print(f'...Features shape: {features.shape}')
            return features, targets, probabilities
        assert features.shape[0] == len(
                targets), '\nFeatures and targets must correspond to the same number of images.\n'
        print(f'...Features successfully extracted for all {len(features)} images in the database.')
        print(f'...Features shape: {features.shape}')
        return features, targets


    def enumerate_layers(self, feature_extractor) -> List[int]:
        layers = []
        k = 0
        for n, m in self.model.named_modules():
            if re.search(r'\d+$', n):
                if isinstance(m, feature_extractor):
                    layers.append(k)
                k += 1
        return layers


    def ensemble_featmaps(
        self,
        activations: dict,
        layers: list,
        pooling: str = 'max',
        alpha: float = 3.,
        beta: float = 5.,
    ) -> torch.Tensor:
        """Concatenate globally (max or average) pooled feature maps."""
        acts = [activations[''.join(('features.', str(l)))] for l in layers]
        func = torch.max if pooling == 'max' else torch.mean
        pooled_acts = [torch.tensor([list(map(func, featmaps))
                                    for featmaps in acts_i]) for acts_i in acts]
        # upweight second-to-last conv layer by 5.
        pooled_acts[-2] = pooled_acts[-2] * alpha
        # upweight last conv layer by 10.
        pooled_acts[-1] = pooled_acts[-1] * beta
        stacked_acts = torch.cat(pooled_acts, dim=1)
        return stacked_acts

    
    def get_activation(self, name):
        """Store copy of hidden unit activations at each layer of model."""
        def hook(model, input, output):
            # store copy of tensor rather than tensor itself
            if isinstance(output, tuple):
                act = output[0]
            else:
                act = output
            try:
                activations[name] = act.clone().detach()
            except AttributeError:
                activations[name] = act.clone()
        return hook


    def register_hook(self):
        """Register a forward hook to store activations."""
        for n, m in self.model.named_modules():
            m.register_forward_hook(self.get_activation(n))
        return self.model


    def get_transformations(self, resize_dim: int = 256, crop_dim: int = 224, apply_center_crop: bool = True):
        if re.search(r'^clip', self.model_name):
            if self.backend != 'pt':
                raise Exception("You need to use PyTorch 'pt' as backend if you want to use the CLIP model.")

            composes = [
                T.Resize(self.clip_n_px, interpolation=Image.BICUBIC)
            ]

            if apply_center_crop:
                composes.append(T.CenterCrop(self.clip_n_px))

            composes += [lambda image: image.convert("RGB"),
                         T.ToTensor(),
                         T.Normalize((0.48145466, 0.4578275, 0.40821073), (0.26862954, 0.26130258, 0.27577711))]

            composition = T.Compose(composes)
            return composition
            
        mean = [0.485, 0.456, 0.406]
        std = [0.229, 0.224, 0.225]
        if self.backend == 'pt':
            normalize = T.Normalize(mean=mean, std=std)
            composes = [T.Resize(resize_dim)]
            if apply_center_crop:
                composes.append(T.CenterCrop(crop_dim))
            composes += [T.ToTensor(), normalize]
            composition = T.Compose(composes)
            return composition
        elif self.backend == 'tf':
            resize_dim = crop_dim
            composes = [layers.experimental.preprocessing.Resizing(resize_dim, resize_dim)]

            if apply_center_crop:
                pass
                #composes.append(layers.experimental.preprocessing.CenterCrop(crop_dim, crop_dim))
            composes += [layers.experimental.preprocessing.Normalization(mean=mean, variance=[std_ * std_ for std_ in std])]
            resize_crop_and_normalize = tf.keras.Sequential(composes)
            return resize_crop_and_normalize





import os 
import unittest

import thingsvision.vision as vision 
import tests.helper as helper 

import numpy as np
import pandas as pd 

class RDMTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        helper.create_test_images()
        model_name = 'vgg16_bn'
        model, _, dl = helper.create_model_and_dl(model_name, 'pt')
        module_name = helper.PT_MODEL_AND_MODULE_NAMES[model_name][0]
        features, _ = model.extract_features(
            data_loader=dl,
            module_name=module_name,
            flatten_acts=False
        )
        for format in helper.FILE_FORMATS:
            # tests whether features can be saved in any of the formats as four-dimensional tensor
            vision.save_features(
                features=features,
                out_path=helper.OUT_PATH,
                file_format=format,
            )


    def test_rdms(self):
        """Test different distance metrics on which RDMs are based."""

        features = np.load(os.path.join(helper.OUT_PATH, 'features.npy'))
        features = features.reshape(helper.NUM_SAMPLES, -1)

        rdms = []
        for distance in helper.DISTANCES:
            rdm = vision.compute_rdm(features, distance)
            self.assertEqual(len(rdm.shape), 2)
            self.assertEqual(features.shape[0], rdm.shape[0], rdm.shape[1])
            vision.plot_rdm(helper.OUT_PATH, features, distance)
            rdms.append(rdm)

        for rdm_i, rdm_j in zip(rdms[:-1], rdms[1:]):
            corr = vision.correlate_rdms(rdm_i, rdm_j)
            self.assertTrue(isinstance(corr, float))
            self.assertTrue(corr > float(-1))
            self.assertTrue(corr < float(1))


class ComparisonTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        helper.create_test_images()

    def test_comparison(self):
        backend = 'pt'
        compare_model_names = ['vgg16_bn', 'vgg19_bn']
        compare_module_names = ['features.23', 'classifier.3']
        corr_mat = vision.compare_models(
            root=helper.TEST_PATH,
            out_path=helper.OUT_PATH,
            model_names=compare_model_names,
            module_names=compare_module_names,
            pretrained=True,
            batch_size=helper.BATCH_SIZE,
            backend=backend,
            flatten_acts=True,
            clip=[False, False],
            save_features=False,
        )
        self.assertTrue(isinstance(corr_mat, pd.DataFrame))
        self.assertEqual(corr_mat.shape, (len(
            compare_model_names), len(compare_module_names)))


class LoadItemsTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        helper.download_item_names()

    def test_item_name_loading(self):
        item_names = vision.load_item_names(helper.DATA_PATH)
        self.assertTrue(isinstance(item_names, np.ndarray))
        self.assertEqual(len(item_names), helper.NUM_OBJECTS)


class FileNamesTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        helper.create_test_images()

    def test_filenames(self):
        import os
        print(os.getcwd())
        file_names = open(os.path.join(helper.OUT_PATH, 'file_names.txt'), 'r').read().split()
        img_files = []
        for root, _, files in os.walk(helper.TEST_PATH):
            for f in files:
                if f.endswith('png'):
                    img_files.append(os.path.join(root, f))
        self.assertEqual(sorted(file_names), sorted(img_files))





import abc 

class Custom(metaclass=abc.ABCMeta):
    def __init__(self, device, backend) -> None:
        self.device = device 
        self.backend = backend

    @abc.abstractmethod
    def create_model(self, backend):
        """ Create the model """


from thingsvision.custom_models.custom import Custom
import torchvision.models as torchvision_models
import torch

class Resnet50_ecoset(Custom):
    def __init__(self, device, backend) -> None:
        super().__init__(device, backend)

    def create_model(self):
        if self.backend == 'pt':
            model = torchvision_models.resnet50(pretrained=False, num_classes=565)
            path_to_weights = 'https://osf.io/abfq4/download'
            state_dict = torch.hub.load_state_dict_from_url(path_to_weights, map_location=self.device, file_name='Resnet50_ecoset')
            model.load_state_dict(state_dict)
            return model


import unittest

import tests.helper as helper 
from thingsvision.model_class import Model
from thingsvision.dataloader import DataLoader

import numpy as np 

class ExtractionPTvsTFTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        helper.create_test_images()
        
    def test_custom_torch_vs_tf_extraction(self):
        values = [2, -10]

        tf_backend = 'tf'
        tf_dataset = helper.SimpleDataset(values, tf_backend)
        tf_dl = DataLoader(
            tf_dataset,
            batch_size=1,
            backend=tf_backend,
        )
        tf_model = Model('VGG16', pretrained=False,
                      device=helper.DEVICE, backend=tf_backend)
        tf_model.model = helper.tf_model

        torch_backend = 'pt'
        pt_dataset = helper.SimpleDataset(values, torch_backend)
        pt_dl = DataLoader(
            pt_dataset,
            batch_size=1,
            backend=torch_backend,
        )
        pt_model = Model('vgg16', pretrained=False,
                      device=helper.DEVICE, backend=torch_backend)
        pt_model.model = helper.pt_model

        layer_name = 'relu'
        tf_features, _ = tf_model.extract_features(
            data_loader=tf_dl,
            module_name=layer_name,
            flatten_acts=False,
            )
        pt_features, _ = pt_model.extract_features(
            data_loader=pt_dl,
            module_name=layer_name,
            flatten_acts=False,
            )
        expected_features = np.array([[2, 2], [0, 0]])
        np.testing.assert_allclose(pt_features, expected_features)
        np.testing.assert_allclose(tf_features, expected_features)

        layer_name = 'relu2'
        tf_features, _ = tf_model.extract_features(
            data_loader=tf_dl,
            module_name=layer_name,
            flatten_acts=False,
            )
        pt_features, _ = pt_model.extract_features(
            data_loader=pt_dl,
            module_name=layer_name,
            flatten_acts=False,
            )
        expected_features = np.array([[4, 4], [0, 0]])
        np.testing.assert_allclose(pt_features, expected_features)
        np.testing.assert_allclose(tf_features, expected_features)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import tensorflow as tf
import torch
import math

from typing import Iterator, Tuple


class DataLoader(object):

    def __init__(
        self,
        dataset: Tuple,
        batch_size: int,
        backend: str,
    ):
        self.dataset = dataset
        self.batch_size = batch_size
        self.backend = backend
        self.n_batches = int(math.ceil(len(self.dataset) / self.batch_size))
        self.remainder = len(self.dataset) % self.batch_size

    def __len__(self) -> int:
        return self.n_batches

    def __iter__(self) -> Iterator:
        return self.get_batches(self.dataset)

    def stack_samples(self, X, y) -> Tuple:
        if self.backend == 'pt':
            X = torch.stack(X, dim=0)
            y = torch.stack(y, dim=0)
        else:
            X = tf.stack(X, axis=0)
            y = tf.stack(y, axis=0)
        return (X, y)

    def get_batches(self, dataset: Tuple) -> Iterator:
        for k in range(self.n_batches):
            if (self.remainder != 0 and k == int(self.n_batches - 1)):
                subset = range(k * self.batch_size, k *
                               self.batch_size + self.remainder)
            else:
                subset = range(k * self.batch_size, (k + 1) * self.batch_size)
            X, y = zip(
                *[dataset[i] for i in subset])
            batch = self.stack_samples(X, y)
            yield batch



import os
import shutil
import unittest

import tests.helper as helper
import thingsvision.vision as vision
import numpy as np 

class FeaturesTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        helper.create_test_images()

    def setUp(self):
        shutil.rmtree(helper.OUT_PATH)
        os.makedirs(helper.OUT_PATH)
        
    def get_2D_features(self):
        model_name = 'vgg16_bn'
        model, _, dl = helper.create_model_and_dl(model_name, 'pt')
        module_name ='classifier.3'
        features, _ = model.extract_features(
            data_loader=dl,
            module_name=module_name,
            flatten_acts=False,
        )
        return features
 
    def get_4D_features(self):
        model_name = 'vgg16_bn'
        model, _, dl = helper.create_model_and_dl(model_name, 'pt')
        module_name ='features.23'
        features, _ = model.extract_features(
            data_loader=dl,
            module_name=module_name,
            flatten_acts=False,
        )
        return features

    def test_postprocessing(self):
        """Test different postprocessing methods (e.g., centering, normalization, compression)."""
        features = self.get_2D_features()
        flattened_features = features.reshape(helper.NUM_SAMPLES, -1)
        centred_features = vision.center_features(flattened_features)
        normalized_features = vision.normalize_features(flattened_features)
        transformed_features = vision.compress_features(
            flattened_features, rnd_seed=42, retained_var=.9)

        # self.assertTrue(centred_features.mean(axis=0).sum() < 1e-3)
        self.assertEqual(np.linalg.norm(normalized_features, axis=1).sum(),
                         np.ones(features.shape[0]).sum())
        self.assertTrue(
            transformed_features.shape[1] < flattened_features.shape[1])

    def check_file_exists(self, file_name, format, txt_should_exists=True):
        if format == 'hdf5':
            format = 'h5'
            file_name = 'features'

        path = os.path.join(helper.OUT_PATH, f'{file_name}.{format}')
        if format == 'txt' and not txt_should_exists:
            self.assertTrue(not os.path.exists(path))
        else:
            self.assertTrue(os.path.exists(path))

    def test_storing_2d(self):
        """Test storing possibilities."""
        features = self.get_2D_features()
        for format in helper.FILE_FORMATS:
            # tests whether features can be saved in any of the formats as two-dimensional tensor
            vision.save_features(
                features=features,
                out_path=helper.OUT_PATH,
                file_format=format,
            )
            
            self.check_file_exists('features', format)

    def test_storing_4d(self):
        features = self.get_4D_features()
        for format in helper.FILE_FORMATS:
            # tests whether features can be saved in any of the formats as four-dimensional tensor
            vision.save_features(
                features=features,
                out_path=helper.OUT_PATH,
                file_format=format,
            )
            
            self.check_file_exists('features', format, False)

    def test_splitting_2d(self):
        n_splits = 3
        features = self.get_2D_features()
        for format in helper.FILE_FORMATS:
            vision.split_features(
                        features=features,
                        PATH=helper.OUT_PATH,
                        file_format=format,
                        n_splits=n_splits
            )
            
            for i in range(1, n_splits):
                self.check_file_exists(f'features_0{i}', format)

    def test_splitting_4d(self):
        n_splits = 3
        features = self.get_4D_features()
        for format in set(helper.FILE_FORMATS) - set(['txt']):
            vision.split_features(
                        features=features,
                        PATH=helper.OUT_PATH,
                        file_format=format,
                        n_splits=n_splits
            )
            
            for i in range(1, n_splits):
                self.check_file_exists(f'features_0{i}', format, False)

        with self.assertRaises(Exception):
            vision.split_features(
                        features=features,
                        PATH=helper.OUT_PATH,
                        file_format='txt',
                        n_splits=n_splits
            )



import unittest

import tests.helper as helper 

import numpy as np 
import tensorflow as tf 
from torchvision import transforms as T
from thingsvision.model_class import Model 

class ModelLoadingTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        helper.create_test_images()
        
    def test_mode_and_device(self):
        model_name = 'vgg16_bn'
        model, dataset, dl = helper.create_model_and_dl(model_name, 'pt')
        self.assertTrue(hasattr(model.model, helper.DEVICE))
        self.assertFalse(model.model.training)

    def test_transformations_clip(self):
        model_name = 'clip-RN'
        model, dataset, dl = helper.create_model_and_dl(model_name, 'pt')
        transforms = model.get_transformations()
        self.assertTrue(isinstance(transforms, T.Compose))

        with self.assertRaisesRegex(Exception, "You need to use PyTorch 'pt' as backend if you want to use the CLIP model."):
            model.backend = 'tf'
            transforms = model.get_transformations()

    def test_transformations_cnn(self):
        model_name = 'vgg16_bn'
        model, dataset, dl = helper.create_model_and_dl(model_name, 'pt')
        transforms = model.get_transformations()
        self.assertTrue(isinstance(transforms, T.Compose))

        model_name = 'VGG16'
        model, dataset, dl = helper.create_model_and_dl(model_name, 'tf')
        transforms = model.get_transformations()
        self.assertTrue(isinstance(transforms, tf.keras.Sequential))

    def test_load_custom_user_model(self):
        model_name = 'VGG16bn_ecoset'
        model = Model(model_name, True, 'cpu')
        self.assertTrue(model.__class__.__name__, 'vgg')

        model_name = 'Resnet50_ecoset'
        model = Model(model_name, True, 'cpu')
        self.assertTrue(model.__class__.__name__, 'resnet')

        model_name = 'Alexnet_ecoset'
        model = Model(model_name, True, 'cpu')
        self.assertTrue(model.__class__.__name__, 'alexnet')




from collections import OrderedDict
import torch
from torch import nn


HASH = '933c001c'


class Flatten(nn.Module):

    """
    Helper module for flattening input tensor to 1-D for the use in Linear modules
    """

    def forward(self, x):
        return x.view(x.size(0), -1)


class Identity(nn.Module):

    """
    Helper module that stores the current tensor. Useful for accessing by name
    """

    def forward(self, x):
        return x


class CORblock_RT(nn.Module):

    def __init__(self, in_channels, out_channels, kernel_size=3, stride=1, out_shape=None):
        super().__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.out_shape = out_shape

        self.conv_input = nn.Conv2d(in_channels, out_channels, kernel_size=kernel_size,
                                    stride=stride, padding=kernel_size // 2)
        self.norm_input = nn.GroupNorm(32, out_channels)
        self.nonlin_input = nn.ReLU(inplace=True)

        self.conv1 = nn.Conv2d(out_channels, out_channels,
                               kernel_size=3, padding=1, bias=False)
        self.norm1 = nn.GroupNorm(32, out_channels)
        self.nonlin1 = nn.ReLU(inplace=True)

        self.output = Identity()  # for an easy access to this block's output

    def forward(self, inp=None, state=None, batch_size=None):
        if inp is None:  # at t=0, there is no input yet except to V1
            inp = torch.zeros([batch_size, self.out_channels, self.out_shape, self.out_shape])
            if self.conv_input.weight.is_cuda:
                inp = inp.cuda()
        else:
            inp = self.conv_input(inp)
            inp = self.norm_input(inp)
            inp = self.nonlin_input(inp)

        if state is None:  # at t=0, state is initialized to 0
            state = 0
        skip = inp + state

        x = self.conv1(skip)
        x = self.norm1(x)
        x = self.nonlin1(x)

        state = self.output(x)
        output = state
        return output, state


class CORnet_RT(nn.Module):

    def __init__(self, times=5):
        super().__init__()
        self.times = times

        self.V1 = CORblock_RT(3, 64, kernel_size=7, stride=4, out_shape=56)
        self.V2 = CORblock_RT(64, 128, stride=2, out_shape=28)
        self.V4 = CORblock_RT(128, 256, stride=2, out_shape=14)
        self.IT = CORblock_RT(256, 512, stride=2, out_shape=7)
        self.decoder = nn.Sequential(OrderedDict([
            ('avgpool', nn.AdaptiveAvgPool2d(1)),
            ('flatten', Flatten()),
            ('linear', nn.Linear(512, 1000))
        ]))

    def forward(self, inp):
        outputs = {'inp': inp}
        states = {}
        blocks = ['inp', 'V1', 'V2', 'V4', 'IT']

        for block in blocks[1:]:
            if block == 'V1':  # at t=0 input to V1 is the image
                this_inp = outputs['inp']
            else:  # at t=0 there is no input yet to V2 and up
                this_inp = None
            new_output, new_state = getattr(self, block)(this_inp, batch_size=len(outputs['inp']))
            outputs[block] = new_output
            states[block] = new_state

        for t in range(1, self.times):
            new_outputs = {'inp': inp}
            for block in blocks[1:]:
                prev_block = blocks[blocks.index(block) - 1]
                prev_output = outputs[prev_block]
                prev_state = states[block]
                new_output, new_state = getattr(self, block)(prev_output, prev_state)
                new_outputs[block] = new_output
                states[block] = new_state
            outputs = new_outputs

        out = self.decoder(outputs['IT'])
        return out



from collections import OrderedDict
from torch import nn


HASH = '5c427c9c'


class Flatten(nn.Module):

    """
    Helper module for flattening input tensor to 1-D for the use in Linear modules
    """

    def forward(self, x):
        return x.view(x.size(0), -1)


class Identity(nn.Module):

    """
    Helper module that stores the current tensor. Useful for accessing by name
    """

    def forward(self, x):
        return x


class CORblock_Z(nn.Module):

    def __init__(self, in_channels, out_channels, kernel_size=3, stride=1):
        super().__init__()
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size=kernel_size,
                              stride=stride, padding=kernel_size // 2)
        self.nonlin = nn.ReLU(inplace=True)
        self.pool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.output = Identity()  # for an easy access to this block's output

    def forward(self, inp):
        x = self.conv(inp)
        x = self.nonlin(x)
        x = self.pool(x)
        x = self.output(x)  # for an easy access to this block's output
        return x


def CORnet_Z():
    model = nn.Sequential(OrderedDict([
        ('V1', CORblock_Z(3, 64, kernel_size=7, stride=2)),
        ('V2', CORblock_Z(64, 128)),
        ('V4', CORblock_Z(128, 256)),
        ('IT', CORblock_Z(256, 512)),
        ('decoder', nn.Sequential(OrderedDict([
            ('avgpool', nn.AdaptiveAvgPool2d(1)),
            ('flatten', Flatten()),
            ('linear', nn.Linear(512, 1000)),
            ('output', Identity())
        ])))
    ]))

    # weight initialization
    for m in model.modules():
        if isinstance(m, (nn.Conv2d, nn.Linear)):
            nn.init.xavier_uniform_(m.weight)
            if m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.BatchNorm2d):
            m.weight.data.fill_(1)
            m.bias.data.zero_()

    return model


from thingsvision.custom_models.resnet50_ecoset import Resnet50_ecoset
from thingsvision.custom_models.vgg16bn_ecoset import VGG16bn_ecoset
from thingsvision.custom_models.alexnet_ecoset import Alexnet_ecoset


import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

requirements = [
                'matplotlib',
                'numba',
                'ftfy',
                'numpy',
                'pandas',
                'tensorflow',
                'torch',
                'torchvision',
                'tqdm',
                'regex',
                'scikit-image',
                'scikit-learn',
                'scipy',
                'h5py',
                ]

setuptools.setup(
                name="thingsvision",
                version="1.4.4",
                author="Lukas Muttenthaler",
                author_email="muttenthaler@cbs.mpg.de",
                description="Extracting image features from state-of-the-art neural networks for Computer Vision made easy",
                long_description=long_description,
                long_description_content_type="text/markdown",
                url="https://github.com/ViCCo-Group/THINGSvision",
                packages=setuptools.find_packages(),
                license="MIT License",
                install_requires=requirements,
                keywords="feature extraction",
                classifiers=[
                    "Programming Language :: Python :: 3.7",
                    "Natural Language :: English",
                    "License :: OSI Approved :: MIT License",
                    "Operating System :: OS Independent",
                ],
                python_requires='>=3.7',
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import h5py
import json
import os
import pickle
import re
import scipy
import scipy.io
import torch
import itertools

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

import thingsvision.cornet as cornet
import thingsvision.clip as clip

import torch.nn as nn
import torch.nn.functional as F
import torchvision.models as models

from collections import defaultdict
from numba import njit, prange
from os.path import join as pjoin
from scipy.stats import rankdata
from scipy.spatial.distance import pdist
from scipy.spatial.distance import squareform
from thingsvision.model_class import Model
from thingsvision.dataset import ImageDataset
from thingsvision.dataloader import DataLoader
from torchvision import transforms as T
from typing import Tuple, List, Iterator, Dict, Any


FILE_FORMATS = ['hdf5', 'npy', 'mat', 'txt']



def load_dl(
    root: str,
    out_path: str,
    backend: str,
    batch_size: int,
    imagenet_train: bool = None,
    imagenet_val: bool = None,
    things: bool = None,
    things_behavior: bool = None,
    add_ref_imgs: bool = None,
    file_names: List[str] = None,
    transforms=None,
) -> Iterator:
    """Create a data loader for custom image dataset

    Parameters
    ----------
    root : str
        Root directory. Directory where images are stored.
    out_path : str
        PATH where order of images features should be stored.
    batch_size : int (optional)
        Number of samples (i.e., images) per mini-batch.
    imagenet_train : bool (optional)
        Whether ImageNet train set is used.
    imagenet_val : bool (optional)
        Whether ImageNet validation set is used.
    things : bool (optional)
        Whether THINGS database is used.
    things_behavior : bool (optional)
        Whether THINGS images used in behavioral experiments
        are used.
    add_ref_imgs : bool (optional)
        Whether the union of the THINGS database and those
        images that were used in behavioral experiments is used.
    file_names : List[str] (optional)
        List of file names. A list of file names that determines
        the order in which image features are extracted can optionally
        be passed.
    transforms : Any
        Composition of image transformations. Must be either a PyTorch composition
        or a Tensorflow Sequential model.

    Returns
    -------
    output : Iterator
        Returns an iterator of image mini-batches.
        Each mini-batch consists of <batch_size> samples.
    """
    print(f'\n...Loading dataset into memory.')
    if not os.path.exists(out_path):
        os.makedirs(out_path)
        print(f'...Creating output directory.')
    dataset = ImageDataset(
        root=root,
        out_path=out_path,
        backend=backend,
        imagenet_train=imagenet_train,
        imagenet_val=imagenet_val,
        things=things,
        things_behavior=things_behavior,
        add_ref_imgs=add_ref_imgs,
        file_names=file_names,
        transforms=transforms,
    )
    print(f'...Transforming dataset into {backend} DataLoader.\n')
    dl = DataLoader(dataset, batch_size=batch_size, backend=backend)
    return dl


def get_module_names(model, module: str) -> list:
    """Extract correct module names, if iterating over multiple modules is desired."""
    if model.backend == 'pt':
        module_names, _ = zip(*model.model.named_modules())
    else:
        module_names = list(map(lambda m: m.name, model.model.layers))    
    return list(filter(lambda n: re.search(f'{module}', n), module_names))


def extract_features_across_models_and_datasets(
    out_path: str,
    model_names: List[str],
    img_paths: List[str],
    module_names: List[str],
    clip: List[bool],
    pretrained: bool,
    batch_size: int,
    backend: str,
    flatten_acts: bool,
    f_format: str = 'txt'
) -> None:
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    for i, model_name in enumerate(model_names):
        model = Model(
            model_name=model_name,
            pretrained=pretrained,
            device=device,
            model_path=None,
            backend=backend,
            )
        transforms = model.get_transformations()
        for img_path in img_paths:
            PATH = os.path.join(out_path, img_path, model_name,
                                module_names[i], 'features')
            dl = load_dl(
                root=img_path,
                out_path=out_path,
                backend=backend,
                batch_size=batch_size,
                transforms=transforms,
                )
            features, _ = model.extract_features(
                data_loader=dl,
                module_name=module_names[i],
                flatten_acts=flatten_acts,
                clip=clip[i],
                )
            save_features(features, PATH, f_format)


def extract_features_across_models_datasets_and_modules(
    out_path: str,
    model_names: List[str],
    img_paths: List[str],
    module_names: List[str],
    clip: List[str],
    pretrained: bool,
    batch_size: int,
    backend: str,
    flatten_acts: bool,
    f_format: str = 'txt'
) -> None:
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    for i, model_name in enumerate(model_names):
        model = Model(
            model_name=model_name,
            pretrained=pretrained,
            device=device,
            model_path=None,
            backend=backend,
            )
        transforms = model.get_transformations()
        modules = get_module_names(model, module_names[i])
        for img_path in img_paths:
            for module_name in modules:
                PATH = os.path.join(out_path, img_path,
                                    model_name, module_name, 'features')
                dl = load_dl(
                    root=img_path,
                    out_path=out_path,
                    backend=backend,
                    batch_size=batch_size,
                    transforms=transforms,
                    )
                features, _ = model.extract_features(
                    data_loader=dl,
                    module_name=module_name,
                    flatten_acts=flatten_acts,
                    clip=clip[i],
                    )
                save_features(features, PATH, f_format)


def center_features(X: np.ndarray) -> np.ndarray:
    """Center features to have zero mean."""
    try:
        X -= X.mean(axis=0)
        return X
    except:
        raise Exception(
            '\nMake sure features are represented through a two-dimensional array\n')


def normalize_features(X: np.ndarray) -> np.ndarray:
    """Normalize feature vectors by their l2-norm."""
    try:
        X /= np.linalg.norm(X, axis=1)[:, np.newaxis]
        return X
    except:
        raise Exception(
            f'\nMake sure features are represented through a two-dimensional array\n')


def compress_features(X: np.ndarray, rnd_seed: int, retained_var: float = .9) -> np.ndarray:
    """Compress feature matrix with Principal Components Analysis (PCA)."""
    from sklearn.decomposition import PCA
    assert isinstance(
        rnd_seed, int), '\nTo reproduce results, random state for PCA must be defined.\n'
    pca = PCA(n_components=retained_var,
              svd_solver='full', random_state=rnd_seed)
    transformed_feats = pca.fit_transform(X)
    return transformed_feats


# ################################################################# #
# HELPER FUNCTIONS FOR SAVING, MERGING AND SLICING FEATURE MATRICES #
# ################################################################# #


def rm_suffix(img: str) -> str:
    return re.sub(r'(.eps|.jpg|.jpeg|.png|.PNG|.tif|.tiff)$', '', img)


def store_features(
    PATH: str,
    features: np.ndarray,
    file_format: str,
) -> None:
    """Save feature matrix to disk in pre-defined file format."""
    if not os.path.exists(PATH):
        print(f'...Output directory did not exist. Creating directories to save features.')
        os.makedirs(PATH)

    if file_format == 'npy':
        with open(pjoin(PATH, 'features.npy'), 'wb') as f:
            np.save(f, features)
    elif file_format == 'mat':
        try:
            with open(pjoin(PATH, 'file_names.txt'), 'r') as f:
                file_names = [rm_suffix(l.strip()) for l in f]
            features = {file_name: feature for file_name,
                        feature in zip(file_names, features)}
            scipy.io.savemat(pjoin(PATH, 'features.mat'), features)
        except FileNotFoundError:
            scipy.io.savemat(pjoin(PATH, 'features.mat'),
                             {'features': features})
    elif file_format == 'hdf5':
        h5f = h5py.File(pjoin(PATH, 'features.h5'), 'w')
        h5f.create_dataset('features', data=features)
        h5f.close()
    else:
        np.savetxt(pjoin(PATH, 'features.txt'), features)
    print(f'...Features successfully saved to disk.\n')


def split_features(
    PATH: str,
    features: np.ndarray,
    file_format: str,
    n_splits: int,
) -> None:
    """Split feature matrix into <n_splits> subsamples to counteract MemoryErrors."""
    if file_format == 'mat':
        try:
            with open(pjoin(PATH, 'file_names.txt'), 'r') as f:
                file_names = [rm_suffix(l.strip()) for l in f]
        except FileNotFoundError:
            file_names = None
    splits = np.linspace(0, len(features), n_splits, dtype=int)
    if file_format == 'hdf5':
        h5f = h5py.File(pjoin(PATH, 'features.h5'), 'w')

    for i in range(1, len(splits)):
        feature_split = features[splits[i - 1]:splits[i]]
        if file_format == 'npy':
            with open(pjoin(PATH, f'features_{i:02d}.npy'), 'wb') as f:
                np.save(f, feature_split)
        elif file_format == 'mat':
            if file_names:
                file_name_split = file_names[splits[i - 1]:splits[i]]
                new_features = {
                    file_name_split[i]: feature for i, feature in enumerate(feature_split)}
                scipy.io.savemat(
                    pjoin(PATH, f'features_{i:02d}.mat'), new_features)
            else:
                scipy.io.savemat(pjoin(PATH, f'features_{i:02d}.mat'), {
                                 'features': features})
        elif file_format == 'hdf5':
            h5f.create_dataset(f'features_{i:02d}', data=feature_split)
        else:
            np.savetxt(pjoin(PATH, f'features_{i:02d}.txt'), feature_split)

    if file_format == 'hdf5':
        h5f.close()


def merge_features(PATH: str, file_format: str) -> np.ndarray:
    if file_format == 'hdf5':
        with h5py.File(pjoin(PATH, 'features.h5'), 'r') as f:
            features = np.vstack([split[:] for split in f.values()])
    else:
        feature_splits = np.array([split for split in os.listdir(PATH) if split.endswith(
            file_format) and re.search(r'^(?=^features)(?=.*[0-9]+$).*$', split.rstrip('.' + file_format))])
        enumerations = np.array([int(re.sub(r'\D', '', feature))
                                for feature in feature_splits])
        feature_splits = feature_splits[np.argsort(enumerations)]
        if file_format == 'txt':
            features = np.vstack([np.loadtxt(pjoin(PATH, feature))
                                 for feature in feature_splits])
        elif file_format == 'mat':
            features = np.vstack([scipy.io.loadmat(pjoin(PATH, feature))['features']
                                 for feature in feature_splits])
        elif file_format == 'npy':
            features = np.vstack([np.load(pjoin(PATH, feature))
                                 for feature in feature_splits])
        else:
            raise Exception(
                '\nCan only process hdf5, npy, mat, or txt files.\n')
    return features


def parse_imagenet_synsets(PATH: str) -> List[str]:
    """Convert WN synsets into classes."""
    def parse_str(str):
        return re.sub(r'[^a-zA-Z]', '', str).rstrip('n').lower()
    imagenet_synsets = []
    with open(PATH, 'r') as f:
        for i, l in enumerate(f):
            l = l.split('_')
            cls = '_'.join(list(map(parse_str, l)))
            imagenet_synsets.append(cls)
    return imagenet_synsets


def parse_imagenet_classes(PATH: str) -> List[str]:
    """Disambiguate ImageNet classes."""
    imagenet_classes = []
    with open(PATH, 'r') as f:
        for i, l in enumerate(f):
            l = l.strip().split()
            cls = '_'.join(l[1:]).rstrip(',').strip("'").lower()
            cls = cls.split(',')
            cls = cls[0]
            imagenet_classes.append(cls)
    return imagenet_classes


def get_class_intersection(imagenet_classes: list, things_objects: list) -> set:
    """Return intersection of THINGS objects and ImageNet classes."""
    return set(things_objects).intersection(set(imagenet_classes))


def get_cls_mapping_imagenet(PATH: str, save_as_json: bool = False) -> dict:
    """Store ImageNet classes in an *index_to_class* dictionary, and subsequently save as .json file."""
    if re.search(r'synset', PATH.split('/')[-1]):
        imagenet_classes = parse_imagenet_synsets(PATH)
    else:
        imagenet_classes = parse_imagenet_classes(PATH)
    idx2cls = dict(enumerate(imagenet_classes))
    if save_as_json:
        filename = 'imagenet_idx2class.json'
        PATH = '/'.join(PATH.split('/')[:-1])
        with open(pjoin(PATH, filename), 'w') as f:
            json.dump(idx2cls, f)
    return idx2cls


def get_class_probabilities(
    probas: np.ndarray,
    out_path: str,
    cls_file: str,
    top_k: int,
    save_as_json: bool,
) -> Dict[str, Dict[str, float]]:
    """Compute probabilities per ImageNet class."""
    file_names = open(pjoin(out_path, 'file_names.txt'),
                      'r').read().splitlines()
    idx2cls = get_cls_mapping_imagenet(cls_file)
    class_probas = {}
    for i, (file, p_i) in enumerate(zip(file_names, probas)):
        sorted_predictions = np.argsort(-p_i)[:top_k]
        class_probas[file] = {idx2cls[pred]: float(
            p_i[pred]) for pred in sorted_predictions}
    if save_as_json:
        with open(pjoin(out_path, 'class_probabilities.json'), 'w') as f:
            json.dump(class_probas, f)
    return class_probas


def json2dict(PATH: str, filename: str) -> dict:
    with open(pjoin(PATH, filename), 'r') as f:
        idx2cls = dict(json.load(f))
    return idx2cls


def compose_transforms(resize_dim: int = 256, crop_dim: int = 224):
    normalize = T.Normalize(mean=[0.485, 0.456, 0.406], std=[
                            0.229, 0.224, 0.225])
    composition = T.Compose(
        [T.Resize(resize_dim), T.CenterCrop(crop_dim), T.ToTensor(), normalize])
    return composition


def load_item_names(folder: str = './data') -> np.ndarray:
    return pd.read_csv(pjoin(folder, 'item_names.tsv'), encoding='utf-8', sep='\t').uniqueID.values


def save_features(
    features: np.ndarray,
    out_path: str,
    file_format: str,
    n_splits: int = 10,
) -> None:
    """Save feature matrix in desired format to disk."""
    assert file_format in FILE_FORMATS, f'\nFile format must be one of {FILE_FORMATS}.\n'
    if not os.path.exists(out_path):
        print(
            f'\nOutput directory did not exist. Creating directories to save features...\n')
        os.makedirs(out_path)
    # save hidden unit actvations to disk (either as one single file or as several splits)
    if len(features.shape) > 2 and file_format == 'txt':
        print(f'\n...Cannot save 4-way tensor in a txt format.')
        print(f'...Change format to one of {FILE_FORMATS[:-1]}.\n')
    else:
        try:
            store_features(PATH=out_path, features=features,
                           file_format=file_format)
        except MemoryError:
            print(
                f'\n...Could not save features as one single file due to memory problems.')
            print(f'...Now splitting features along row axis into several batches.\n')
            split_features(PATH=out_path, features=features,
                           file_format=file_format, n_splits=n_splits)
            print(
                f'...Saved features in {n_splits:02d} different files, enumerated in ascending order.')
            print(f'If you want features to be splitted into more or fewer files, simply change number of splits parameter.\n')


def save_targets(
    targets: np.ndarray,
    PATH: str,
    file_format: str,
) -> None:
    """Save target vector to disk."""
    if not os.path.exists(PATH):
        print(
            f'\nOutput directory did not exist. Creating directories to save targets...\n')
        os.makedirs(PATH)

    if file_format == 'npy':
        with open(pjoin(PATH, 'targets.npy'), 'wb') as f:
            np.save(f, targets)
    elif file_format == 'mat':
        scipy.io.savemat(pjoin(PATH, 'targets.mat'), {'targets': targets})
    elif file_format == 'hdf5':
        h5f = h5py.File(pjoin(PATH, 'targets.h5'), 'w')
        h5f.create_dataset('targets', data=targets)
        h5f.close()
    else:
        np.savetxt(pjoin(PATH, 'targets.txt'), targets)
    print(f'...Targets successfully saved to disk.\n')

# ########################################### #
# HELPER FUNCTIONS FOR RSA & RDM COMPUTATIONS #
# ########################################### #


@njit(parallel=True, fastmath=True)
def squared_dists(F: np.ndarray) -> np.ndarray:
    """Compute squared l2-distances between feature representations in parallel."""
    N = F.shape[0]
    D = np.zeros((N, N))
    for i in prange(N):
        for j in prange(N):
            D[i, j] = np.linalg.norm(F[i] - F[j]) ** 2
    return D


def gaussian_kernel(F: np.ndarray) -> np.ndarray:
    """Compute dissimilarity matrix based on a Gaussian kernel."""
    D = squared_dists(F)
    return np.exp(-D / np.mean(D))


def correlation_matrix(F: np.ndarray, a_min: float = -1., a_max: float = 1.) -> np.ndarray:
    """Compute dissimilarity matrix based on correlation distance (on the matrix-level)."""
    F_c = F - F.mean(axis=1)[:, np.newaxis]
    cov = F_c @ F_c.T
    # compute vector l2-norm across rows
    l2_norms = np.linalg.norm(F_c, axis=1)
    denom = np.outer(l2_norms, l2_norms)
    corr_mat = (cov / denom).clip(min=a_min, max=a_max)
    return corr_mat


def cosine_matrix(F: np.ndarray, a_min: float = -1., a_max: float = 1.) -> np.ndarray:
    """Compute dissimilarity matrix based on cosine distance (on the matrix-level)."""
    num = F @ F.T
    # compute vector l2-norm across rows
    l2_norms = np.linalg.norm(F, axis=1)
    denom = np.outer(l2_norms, l2_norms)
    cos_mat = (num / denom).clip(min=a_min, max=a_max)
    return cos_mat


def compute_rdm(F: np.ndarray, method: str) -> np.ndarray:
    """Compute representational dissimilarity matrix based on some distance measure.

    Parameters
    ----------
    F : ndarray
        Input array. Feature matrix of size n x m,
        where n corresponds to the number of observations
        and m is the number of latent dimensions.
    method : str
        Distance metric (e.g., correlation, cosine).

    Returns
    -------
    output : ndarray
        Returns the representational dissimilarity matrix.
    """
    methods = ['correlation', 'cosine', 'euclidean', 'gaussian']
    assert method in methods, f'\nMethod to compute RDM must be one of {methods}.\n'
    if method == 'euclidean':
        rdm = squareform(pdist(F, method))
        return rdm
    else:
        if method == 'correlation':
            rsm = correlation_matrix(F)
        elif method == 'cosine':
            rsm = cosine_matrix(F)
        elif method == 'gaussian':
            rsm = gaussian_kernel(F)
    return 1 - rsm


def correlate_rdms(
    rdm_1: np.ndarray,
    rdm_2: np.ndarray,
    correlation: str = 'pearson',
) -> float:
    """Correlate the upper triangular parts of two distinct RDMs.

    Parameters
    ----------
    rdm_1 : ndarray
        First RDM.
    rdm_2 : ndarray
        Second RDM.
    correlation : str
        Correlation coefficient (e.g., Spearman, Pearson).

    Returns
    -------
    output : float
        Returns the correlation coefficient of the two RDMs.
    """
    triu_inds = np.triu_indices(len(rdm_1), k=1)
    corr_func = getattr(scipy.stats, ''.join((correlation, 'r')))
    rho = corr_func(rdm_1[triu_inds], rdm_2[triu_inds])[0]
    return rho


def plot_rdm(
    out_path: str,
    F: np.ndarray,
    method: str = 'correlation',
    format: str = '.png',
    colormap: str = 'cividis',
    show_plot: bool = False,
) -> None:
    """Compute and plot representational dissimilarity matrix based on some distance measure.

    Parameters
    ----------
    out_path : str
        Output directory. Directory where to store plots.
    F : ndarray
        Input array. Feature matrix of size n x m,
        where n corresponds to the number of observations
        and m is the number of latent dimensions.
    method : str
        Distance metric (e.g., correlation, cosine).
    format : str
        Image format in which to store visualized RDM.
    colormap : str
        Colormap for visualization of RDM.
    show_plot : bool
        Whether to show visualization of RDM after storing it to disk.

    Returns
    -------
    output : ndarray
        Returns the representational dissimilarity matrix.
    """
    rdm = compute_rdm(F, method)
    plt.figure(figsize=(10, 4), dpi=200)
    plt.imshow(rankdata(rdm).reshape(rdm.shape),
               cmap=getattr(plt.cm, colormap))
    plt.xticks([])
    plt.yticks([])
    plt.tight_layout()
    if not os.path.exists(out_path):
        print(f'\n...Output directory did not exists. Creating directories.\n')
        os.makedirs(out_path)
    plt.savefig(os.path.join(out_path, ''.join(('rdm', format))))
    if show_plot:
        plt.show()
    plt.close()


def get_features(
    root: str,
    out_path: str,
    model_names: List[str],
    module_names: List[str],
    clip: List[bool],
    pretrained: bool,
    batch_size: int,
    backend: str,
    flatten_acts: bool,
) -> Dict[str, Dict[str, np.ndarray]]:
    """Extract features for a list of neural network models and corresponding modules.

    Parameters
    ----------
    root : str
        Root directory. Directory where images are stored.
    out_path : str
        PATH where order of images features should be stored.
        Files are alphabetically sorted and features are
        extracted accordingly.
    model_names : List[str]
        List of neural network models for which features
        should be extracted.
    module_names : List[str]
        List of neural network layers for which features
        should be extracted. Modules must correspond to
        models. This should be thought of as zipped lists.
    clip : List[bool]
        List of Booleans which indicates whether the
        corresponding model in the <model_names> list
        is a CLIP-based model or not (i.e., True if
        CLIP, else False)
    pretrained : bool
        Whether pretrained or randomly initialized models
        should be loaded into memory.
    batch_size : int
        Integer value that determines the number of images
        within a single mini-batch (i.e., subsample
        of the data).
    flatten_acts : bool
        Whether activation tensor (e.g., activations
        from an early layer of the neural network model)
        should be transformed into a feature vector.

    Returns
    -------
    output : Dict[str, Dict[str, np.ndarray]]
        Returns a dictionary of feature matrices
        corresponding to the selected models and layers.
    """
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    model_features = defaultdict(dict)
    for i, model_name in enumerate(model_names):
        model = Model(
            model_name=model_name,
            pretrained=pretrained,
            device=device,
            model_path=None,
            backend=backend,
            )
        transforms = model.get_transformations()
        dl = load_dl(
                    root=root,
                    out_path=out_path,
                    backend=backend,
                    batch_size=batch_size,
                    transforms=transforms,
                    )
        features, _ = model.extract_features(
            data_loader=dl,
            module_name=module_names[i],
            flatten_acts=flatten_acts,
            clip=clip[i],
            )
        model_features[model_name][module_names[i]] = features
    return model_features


def compare_models(
    root: str,
    out_path: str,
    model_names: List[str],
    module_names: List[str],
    pretrained: bool,
    batch_size: int,
    backend: str,
    flatten_acts: bool,
    clip: List[bool],
    save_features: bool = True,
    dissimilarity: str = 'correlation',
    correlation: str = 'pearson',
) -> pd.DataFrame:
    """Compare object representations of different models against each other.

    Parameters
    ----------
    root : str
        Root directory. Directory from where to load images.
    out_path : str
        Output directory. Directory where to store features
        corresponding to each neural network model.
    model_names : List[str]
        List of neural network models whose object representations
        should be compared against.
    module_names : List[str]
        List of neural network layers for which features
        should be extracted. Modules must correspond to
        models. This should be thought of as zipped lists.
    pretrained : bool
        Whether pretrained or randomly initialized models
        should be loaded into memory.
    batch_size : int
        Integer value that determines the number of images
        within a single mini-batch (i.e., subsample
        of the data).
    flatten_acts : bool
        Whether activation tensor (e.g., activations
        from an early layer of the neural network model)
        should be transformed into a feature vector.
    clip : List[bool]
        List of Booleans which indicates whether the
        corresponding model in the <model_names> list
        is a CLIP-based model or not (i.e., True if
        CLIP, else False)
    save_features : bool
        Whether to save model features or solely compare
        their representations against each other
        without saving the features to disk.
    dissimilarity : str
        Distance metric to be used to compute RDMs
        corresponding to the model features.
    correlation : str
        Correlation coefficient (e.g., Spearman or Pearson)
        to be used when performing RSA.

    Returns
    -------
    output : pd.DataFrame
        Returns a correlation matrix whose rows and columns
        correspond to the names of the models in <model_names>.
        The cell elements are the correlation coefficients
        for each model combination. The dataframe can subsequently
        be converted into a heatmap with matplotlib or seaborn.
    """
    # extract features for each model and corresponding module
    model_features = get_features(
        root=root,
        out_path=out_path,
        model_names=model_names,
        module_names=module_names,
        clip=clip,
        pretrained=pretrained,
        batch_size=batch_size,
        backend=backend,
        flatten_acts=flatten_acts,
    )
    # save model features to disc
    if save_features:
        pickle_file_(model_features, out_path, 'features')
    # compare features of each model combination for N bootstraps
    corrs = pd.DataFrame(np.eye(len(model_names)), index=np.arange(
        len(model_names)), columns=model_names, dtype=float)
    model_combs = list(itertools.combinations(model_names, 2))
    for (model_i, model_j) in model_combs:
        module_i = module_names[model_names.index(model_i)]
        module_j = module_names[model_names.index(model_j)]
        features_i = model_features[model_i][module_i]
        features_j = model_features[model_j][module_j]
        rdm_i = compute_rdm(features_i, dissimilarity)
        rdm_j = compute_rdm(features_j, dissimilarity)
        corr = correlate_rdms(rdm_i, rdm_j, correlation)
        corrs.loc[model_names.index(model_i), model_j] = corr
        corrs.loc[model_names.index(model_j), model_i] = corr
    corrs['model_names'] = corrs.columns.to_list()
    corrs.set_index('model_names', inplace=True, drop=True)
    return corrs


def pickle_file_(file: dict, out_path: str, f_name: str) -> None:
    """Pickle any file."""
    with open(os.path.join(out_path, f_name + '.p'), 'wb') as f:
        pickle.dump(file, f)



import re
import unittest
from tests import helper  
from thingsvision.dataloader import DataLoader
from thingsvision.model_class import Model

import numpy as np 

class ExtractionCustomModelTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        helper.create_test_images()
        
    def test_extract_features(self):
        layer_name = 'relu'
        values = [2, -10]
        backends = [['pt', helper.pt_model, 'vgg16'], ['tf', helper.tf_model, 'VGG16']]
        batch_size = 1
        for backend, custom_model, vgg_model in backends:
            dataset = helper.SimpleDataset(values, backend)
            dl = DataLoader(
                dataset,
                batch_size=batch_size,
                backend=backend,
            )
            model = Model(vgg_model, pretrained=False,
                        device=helper.DEVICE, backend=backend)

            model.model = custom_model
            expected_features = np.array([[2, 2], [0, 0]])
            expected_targets = np.array([0, 0])

            features, targets = model.extract_features(
                data_loader=dl, 
                module_name=layer_name, 
                flatten_acts=False, 
                return_probabilities=False)
            np.testing.assert_allclose(features, expected_features)
            np.testing.assert_allclose(targets, expected_targets)


            expected_probs = np.array([[0.5, 0.5], [0.5, 0.5]])
            features, targets, probs = model.extract_features(
                data_loader=dl, 
                module_name=layer_name,
                flatten_acts=False,
                return_probabilities=True)
            np.testing.assert_allclose(features, expected_features)
            np.testing.assert_allclose(targets, expected_targets)
            np.testing.assert_allclose(probs, expected_probs)

    def test_extraction_batches(self):
        values = [1] * 10
        backend = 'pt'
        dataset = helper.SimpleDataset(values, backend)
        model = Model('vgg16', pretrained=False, device=helper.DEVICE, backend=backend)
        model.model = helper.pt_model

        # no batch remainders -> 5 batches with 2 examples
        # batch remainders -> 3 batches with 3 examples and 1 batch with 1 remainder

        for batch_size in [2, 3]:
            dl = DataLoader(
                    dataset,
                    batch_size=batch_size,
                    backend=backend,
            )
            features, targets = model.extract_features(
                data_loader=dl, 
                module_name='relu',
                flatten_acts=False, 
                return_probabilities=False)
            self.assertEqual(features.shape[0], len(dataset))
            self.assertEqual(targets.shape[0], len(dataset))
       
        



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import re

import torch

import numpy as np
import pandas as pd
import tensorflow as tf
from torchvision import transforms as T

import thingsvision.vision as vision

from collections import defaultdict
from os.path import join as pjoin

from PIL import Image
from typing import Tuple, List, Dict, Any


class ImageDataset(object):
    """Generic image dataset class

    Parameters
    ----------
    root : str
        Root directory. Directory from where to load images.
    out_path : str
        PATH where order of images features should be stored.
    imagenet_train : bool (optional)
        Whether ImageNet train set is used.
    imagenet_val : bool (optional)
        Whether ImageNet validation set is used.
    things : bool (optional)
        Whether THINGS database is used.
    things_behavior : bool (optional)
        Whether THINGS images used in behavioral experiments
        are used.
    add_ref_imgs : bool (optional)
        Whether the union of the THINGS database and those
        images that were used in behavioral experiments is used.
    file_names : List[str] (optional)
        List of file names. A list of file names that determines
        the order in which image features are extracted can optionally
        be passed.
    transforms : Any
        Composition of image transformations. Must be either a PyTorch composition
        or a Tensorflow Sequential model.

    Returns
    -------
    output : Dataset[Tuple[torch.Tensor, torch.Tensor]]
        Returns a generic image dataset of instance
        (i.e., image in matrix format) and target
        (i.e., class label) tensors.
    """

    def __init__(
        self,
        root: str,
        out_path: str,
        backend: str,
        imagenet_train: bool,
        imagenet_val: bool,
        things: bool,
        things_behavior: bool,
        add_ref_imgs: bool,
        file_names: List[str] = None,
        transforms=None,
    ):
        self.root = root
        self.backend = backend
        self.imagenet_train = imagenet_train
        self.imagenet_val = imagenet_val
        self.things = things
        self.things_behavior = things_behavior
        self.transforms = transforms
        self.file_names = file_names
        self.backend = backend

        classes, idx_to_cls, cls_to_idx, class_folders = self.find_classes_()
        self.class_dataset = class_folders

        if self.imagenet_val:
            img2cls, synsets = get_img2cls_mapping(self.root)
            self.classes = synsets

        if self.class_dataset:
            if self.file_names:
                _, cls_to_files = get_classes(self.file_names)
                self.samples = class_dataset(
                    self.root, out_path, cls_to_idx, self.imagenet_train, self.things, add_ref_imgs, cls_to_files)
            else:
                self.samples = class_dataset(
                    self.root, out_path, cls_to_idx, self.imagenet_train, self.things, add_ref_imgs)
            self.classes = classes
        else:
            self.samples = instance_dataset(
                self.root, out_path, classes, self.imagenet_val, img2cls if self.imagenet_val else None)
            if not self.imagenet_val:
                self.classes = list(cls_to_idx.keys())

        images, targets = zip(*self.samples)

        self.idx_to_cls = idx_to_cls
        self.cls_to_idx = cls_to_idx
        self.targets = targets
        self.images = images

    def find_classes_(self) -> Tuple[list, dict, dict]:
        children = sorted(
            [d.name for d in os.scandir(self.root) if d.is_dir()])
        if children:
            class_folders = True
            if self.file_names:
                classes, _ = get_classes(self.file_names)
            else:
                if self.things:
                    data_path = './data'
                    concept_file = 'things_concepts.tsv'
                    if not os.path.exists(os.path.join(data_path, concept_file)):
                        os.mkdir(data_path)
                        print(f'\n...Created PATH: {data_path}\n')
                        raise FileNotFoundError(
                            f'To extract features for THINGS images, concept file is required. Move {concept_file} to {data_path}.')

                    concept_ids = pd.read_csv(pjoin(data_path, concept_file),
                                              encoding='utf-8', sep='\t').uniqueID.tolist()
                    assert len(children) == len(
                        concept_ids), '\nNumber of categories in dataset must be equal to the number of concept IDs. Check img folder.\n'
                    classes = children if children == concept_ids else concept_ids
                else:
                    classes = children
            if self.imagenet_train:
                cls_to_idx = get_syn2cls_mapping(self.root)
                idx_to_cls = {idx: cls for cls, idx in cls_to_idx.items()}
                classes = list(cls_to_idx.keys())
            else:
                idx_to_cls = dict(enumerate(classes))
        else:
            class_folders = False
            if self.things_behavior:
                # sort objects according to item names in THINGS database
                classes = [''.join((name, '.jpg'))
                           for name in vision.load_item_names()]
            else:
                if self.file_names:
                    classes = list(filter(parse_img_name, self.file_names))
                else:
                    classes = sorted([f.name for f in os.scandir(self.root)
                                     if f.is_file() and parse_img_name(f.name)])
            idx_to_cls = dict(enumerate(list(map(rm_suffix, classes))))
        if not self.imagenet_train:
            cls_to_idx = {cls: idx for idx, cls in idx_to_cls.items()}
        return classes, idx_to_cls, cls_to_idx, class_folders

    def __getitem__(self, idx: int) -> Tuple[Any, Any]:
        img_path, target = self.samples[idx]
        img = Image.open(img_path).convert('RGB')
        img, target = self.transform_img_target(img, target)
            
        return img, target

    def transform_img_target(self, img, target):
        if self.transforms:
            if self.backend == 'pt':
                img = self.transforms(img)
                target = torch.tensor([target])
            elif self.backend == 'tf':
                img = tf.keras.preprocessing.image.img_to_array(img)
                img = self.transforms(img)
                target = tf.convert_to_tensor(target)
        else:
            target = np.array([target])
        return img, target

    def __len__(self) -> int:
        return len(self.samples)


def get_syn2cls_mapping(root: str) -> Dict[str, int]:
    """Maps ImageNet classes (i.e., Wordnet synsets) to numeric labels."""
    try:
        with open(os.path.join(root, 'LOC_synset_mapping.txt'), 'r') as f:
            cls2syn = dict(
                enumerate(list(map(lambda x: x.split()[0], f.readlines()))))
    except FileNotFoundError:
        raise Exception(
            '\nCould not find LOC_synset_mapping.txt in root directory. Move file to root directory.\n')
    syn2cls = {syn: cls for cls, syn in cls2syn.items()}
    return syn2cls


def get_img2cls_mapping(root: str) -> Dict[str, int]:
    """Maps each image in the ImageNet subsample to its corresponding class."""
    split = root.split('/')[-1] if root.split('/')[-1] else root.split('/')[-2]
    try:
        solution = pd.read_csv(os.path.join(root, f'LOC_{split}_solution.csv'))
    except FileNotFoundError:
        raise Exception(
            f'\nCould not find LOC_{split}_solution.csv in root directory. Move file to root directory.\n')
    syn2cls = get_syn2cls_mapping(root)
    img2cls = {row[0] + '.JPEG': syn2cls[row[1].split()[0]]
               for idx, row in solution.iterrows()}
    img2cls = dict(sorted(img2cls.items(), key=lambda kv: kv[0]))
    return img2cls, list(syn2cls.keys())


def get_classes(file_names: List[str]) -> Tuple[List[str], Dict[str, list]]:
    cls_to_files = defaultdict(list)
    classes = []
    for file in file_names:
        if re.search(r'\\', file):
            cls, f_name = file.split('\\')
        elif re.search(r'(/|//)', file):
            cls, f_name = file.split('/')
        else:
            continue
        if cls not in classes:
            classes.append(cls)
        cls_to_files[cls].append(f_name)
    return classes, cls_to_files


def parse_img_name(img_name: str) -> bool:
    return re.search(r'(.eps|.jpg|.JPG|.jpeg|.JPEG|.png|.PNG|.tif|.tiff)$', img_name)


def rm_suffix(img: str) -> str:
    return re.sub(r'(.eps|.jpg|.JPG|.jpeg|.JPEG|.png|.PNG|.tif|.tiff)$', '', img)


def instance_dataset(
    root: str,
    out_path: str,
    images: list,
    imagenet: bool = False,
    img2cls: dict = None,
) -> List[Tuple[str, int]]:
    """Creates a custom <instance> image dataset of image and class label pairs."""
    instances = []
    with open(os.path.join(out_path, 'file_names.txt'), 'w') as f:
        for img in images:
            f.write(f'{img}\n')
            instances.append(os.path.join(root, img))
    if imagenet:
        assert isinstance(
            img2cls, dict), '\nImage-to-cls mapping must be provided\n'
        samples = tuple(
            (instance, img2cls[instance.split('/')[-1]]) for instance in instances)
    else:
        samples = tuple((instance, target)
                        for target, instance in enumerate(instances))
    return samples


def class_dataset(
    PATH: str,
    out_path: str,
    cls_to_idx: Dict[str, int],
    imagenet: bool = None,
    things: bool = None,
    add_ref_imgs: bool = None,
    cls_to_files: Dict[str, List[str]] = None,
) -> List[Tuple[str, int]]:
    """Creates a custom <class> image dataset of image and class label pairs.

    Parameters
    ----------
        PATH : str
            Parent directory. Directory from where to load images.
        out_path : str
            PATH where order of images features should be stored.
        cls_to_idx : Dict[str, int]
            Dictionary of class to numeric label mapping.
        imagenet : bool (optional)
            Whether a subset of ImageNet is used.
        things : bool (optional)
            Whether the THINGS database is used.
        add_ref_imgs : bool (optional)
            Whether the union of the THINGS database and those
            images that were used in behavioral experiments is used.
        cls_to_files : Dict[str, List[str]] (optional)
            Dictionary that maps each class to a list of file names.
            For each class, the list of file names determines
            the order in which image features are extracted.

    Returns
    -------
    output : List[Tuple[str, int]]
        Returns a list of image and class label pairs.
    """
    samples = []
    with open(os.path.join(out_path, 'file_names.txt'), 'w') as f:
        classes = cls_to_idx.keys() if (things or imagenet) else sorted(cls_to_idx.keys())
        for target_cls in classes:
            cls_idx = cls_to_idx[target_cls]
            target_dir = os.path.join(PATH, target_cls)
            if os.path.isdir(target_dir):
                if cls_to_files is None:
                    for root, _, files in sorted(os.walk(target_dir, followlinks=True)):
                        if (things and add_ref_imgs):
                            first_img = sorted(files)[0]
                            if parse_img_name(first_img):
                                first_img = first_img.rstrip('.jpg')
                                if not first_img.endswith('b'):
                                    ref_img_path = get_ref_img(first_img)
                                    item = ref_img_path, cls_idx
                                    samples.append(item)
                                    f.write(f'{ref_img_path}\n')
                            else:
                                raise Exception(
                                    f'\nFound file that does not seem to be in the correct format: {first_img}.\nRemove file before proceeding with feature extraction.\n')
                        for file in sorted(files):
                            path = os.path.join(root, file)
                            if os.path.isfile(path) and parse_img_name(file):
                                item = path, cls_idx
                                samples.append(item)
                                f.write(f'{path}\n')
                else:
                    for f_name in cls_to_files[target_cls]:
                        path = os.path.join(target_dir, f_name)
                        if os.path.isfile(path) and parse_img_name(f_name):
                            item = path, cls_idx
                            samples.append(item)
                            f.write(f'{path}\n')
    return samples


def get_ref_img(
    first_img: str,
    folder: str = './reference_images/',
    suffix: str = '.jpg',
) -> str:
    """Create union of THINGS database and images used in behavioral experiments."""
    if not os.path.exists(folder):
        raise FileNotFoundError(
            f'Directory for reference images not found. Move reference images to {folder}.')
    ref_images = [f.name for f in os.scandir(
        folder) if f.is_file() and parse_img_name(f.name)]
    for ref_img in ref_images:
        img_name = ref_img.rstrip(suffix)
        if re.search(f'^{img_name}', first_img):
            return os.path.join(folder, ref_img)
